==> pftpd-0.33.tar.gz: pftpd-0.33/src/ftpdata.c <==
/*
** ftpdata.c - Manage the data connection
**
** Copyright (c) 1999-2000 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "config.h"

#include <stdio.h>
#include <syslog.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "pftpd.h"

#include "plib/aalloc.h"
#include "plib/threads.h"
#include "plib/safeio.h"
#include "plib/safestr.h"
#include "plib/support.h"
#include "plib/timeout.h"

#define PASV_TIMEOUT 120

static void *
ftpdata_thread(void *vp)
{
    FTPDATA *fdata = (FTPDATA *) vp;
    int err;
    


    if (debug)
	fprintf(stderr, "ftpdata_thread: start\n");
    
    err = fdata->handler(fdata->fp, fdata->vp);
    
    pthread_mutex_lock(&fdata->fp->data_mtx);
    s_close(fdata->fp->data_fd);
    fdata->fp->data_fd = -1;
    pthread_mutex_unlock(&fdata->fp->data_mtx);
    pthread_cond_broadcast(&fdata->fp->data_cv);
    
    switch (err)
    {
      case 0:
	break;
	
      case 226:
	fd_puts(fdata->fp->fd, "226 Transfer complete.\n");
	break;

      case 425:
	fd_puts(fdata->fp->fd, "425 Can not build data connection.\n");
	break;

      case 426:
	fd_puts(fdata->fp->fd, "426 Connection closed; transfer aborted.\n");
	break;

      default:
	fd_printf(fdata->fp->fd, "%d ???\n");
    }
    
    a_free(fdata);
    if (debug)
	fprintf(stderr, "ftpdata_thread: Stop\n");
    
    return NULL;

}


#if 0
static void
timeout_handler(void *vp)
{
    int fd = * (int *) vp;

    write(1, "TIMEOUT\n", 8);

    /* XXX: This is not good, what happens if there is a race between
       this code and s_accept(). Fixme! */
    
    * (int *) vp = -1;
    s_close(fd);
}
#endif


int
ftpdata_run(FTPCLIENT *fp,
	    const char *reason,
	    int (*handler)(FTPCLIENT *fp, void *vp),
	    void *vp)
{
    int err, try;
    int one = 1;
    FTPDATA *fdata;
    pthread_t tid;
    struct sockaddr_gen sin;
    socklen_t slen;
    pthread_attr_t cattr_detached;
    
    
    if (debug)
	fprintf(stderr, "ftpdata_run(\"%s\"): start\n",
		reason);
    
    /* Already have an active FTPDATA session? -> Fail */
    pthread_mutex_lock(&fp->data_mtx);
    if (fp->data_fd != -1)
    {
	pthread_mutex_unlock(&fp->data_mtx);
	syslog(LOG_ERR, "Multiple active FTPDATA sessions not supported");
	return -1;
    }
    

    fd_printf(fp->fd,
	      "150 Opening %s mode data connection for %s.\n",
	      fp->type == ftp_binary ? "BINARY" : "ASCII",
	      reason);

    if (fp->pasv != -1)
    {
	/* Passive connection */

	if (debug)
	    fprintf(stderr, "Waiting for PASV connection\n");
	
	slen = sizeof(sin);
	fp->data_fd = s_accept2(fp->pasv, (struct sockaddr *) &sin, &slen,
				PASV_TIMEOUT);
	pthread_mutex_unlock(&fp->data_mtx);
	
	if (fp->data_fd < 0)
	{
	    if (fp->pasv != -1)
	    {
		s_close(fp->pasv);
		fp->pasv = -1;
	    }
	    return -1;
	    
	    syslog(LOG_ERR,
		   "accept() failed when waiting for Passive FTPDATA connection");
	}

	
	if (debug)
	{
	    char buf1[256];
	    
	    fprintf(stderr, "Got passive connection from:\n");
	    fprintf(stderr, "\tsin_family = %d\n", SGFAM(sin));
	    fprintf(stderr, "\tsin_addr:sin_port = %s\n",
		    s_inet_ntox(&sin, buf1, sizeof(buf1)));
	}
	
	/* Make sure the connection came from the same host as the
	   control connection was from */

	if (SGSIZE(sin) != SGSIZE(fp->cp->rsin) ||
	    memcmp(SGADDRP(sin),
		   SGADDRP(fp->cp->rsin),
		   SGSIZE(sin)) != 0)
	{
	    char buf1[256], buf2[256];
	    
	    syslog(LOG_WARNING,
		   "%s - received foreign PASV connection from %s\n",
		   s_inet_ntox(&fp->cp->rsin, buf1, sizeof(buf1)),
		   s_inet_ntox(&sin, buf2, sizeof(buf2)));

	    pthread_mutex_unlock(&fp->data_mtx);
	    s_close(fp->data_fd);
	    fp->data_fd = -1;
	    pthread_mutex_unlock(&fp->data_mtx);
	    fd_puts(fp->fd, "426 Connection closed; transfer aborted.\n");
	    return 0;
	}

	s_close(fp->pasv);
	fp->pasv = -1;
    }
    else
    {
	/* Active connection */
	
	if (!SGPORT(fp->port))
	{
	    pthread_mutex_unlock(&fp->data_mtx);
	    syslog(LOG_ERR, "Sanity check - result: Total insanity");
	    return -1;
	}
	
	SGINIT(sin);
	slen = SGSOCKSIZE(sin);

	/* Get the local address that the FTP client connected to */
	if (getsockname(fp->cp->fd, (struct sockaddr *) &sin, &slen) != 0)
	{
	    pthread_mutex_unlock(&fp->data_mtx);
	    syslog(LOG_WARNING, "getsockname: %m");
	    return -1;
	}
	
	fp->data_fd = socket(SGFAM(sin), SOCK_STREAM, 0);
	pthread_mutex_unlock(&fp->data_mtx);
	
	if (fp->data_fd < 0)
	{
	    syslog(LOG_ERR, "Failed allocating socket() for Active FTPDATA session");
	    return -1;
	}
	
	if (setsockopt(fp->data_fd,
		       SOL_SOCKET, SO_REUSEADDR, (void *) &one, sizeof(one)))
	{
	    syslog(LOG_WARNING, "setsockopt(SO_REUSEADDR): %m");
	}
	
	if (debug)
	{
	    char buf1[256];

	    
	    fprintf(stderr, "ftpdata_run(): setsockopts done\n");
	    fprintf(stderr, "sin_family = %d\n", SGFAM(sin));
	    fprintf(stderr, "sin_addr:sin_port = %s\n",
		    s_inet_ntox(&sin, buf1, sizeof(buf1)));
	}

	SGPORT(sin) = 0; /* Any available source port */
	
	try = 0;
	while (s_bind(fp->data_fd, (struct sockaddr *) &sin, SGSOCKSIZE(sin)) < 0)
	{
	    char buf1[256];
	    
	    if (debug)
		fprintf(stderr,
			"ftpdata_run: s_bind(): try=%d, addr=%s, failed: %s\n",
			try,
			s_inet_ntox(&sin, buf1, sizeof(buf1)),
			strerror(errno));
	    
	    if (try++ > 10 || errno != EADDRINUSE)
	    {
		char buf1[256];

		
		syslog(LOG_WARNING, "bind(%d,%s): %m",
		       fp->data_fd,
		       s_inet_ntox(&fp->port, buf1, sizeof(buf1)));
		
		pthread_mutex_lock(&fp->data_mtx);
		s_close(fp->data_fd);
		fp->data_fd = -1;
		pthread_mutex_unlock(&fp->data_mtx);
		return -1;
	    }
	    sleep(try);
	}
	
	if (connect(fp->data_fd,
		    (struct sockaddr *) &fp->port,
		    SGSOCKSIZE(fp->port)) < 0)
	{
	    char buf1[256];

	    
	    syslog(LOG_WARNING, "connect(%d,%s): %m",
		   fp->data_fd,
		   s_inet_ntox(&fp->port, buf1, sizeof(buf1)));
	    
	    pthread_mutex_lock(&fp->data_mtx);
	    s_close(fp->data_fd);
	    fp->data_fd = -1;
	    pthread_mutex_unlock(&fp->data_mtx);
	    return -1;
	}

    }
    
    if (setsockopt(fp->data_fd,
		   IPPROTO_TCP, TCP_NODELAY, (void *) &one, sizeof(one)))
    {
	syslog(LOG_WARNING, "setsockopt(TCP_NODELAY): %m");
    }
    
#ifdef IPTOS_THROUGHPUT
    {
	int tos = IPTOS_THROUGHPUT;
	if (setsockopt(fp->data_fd,
		       IPPROTO_IP, IP_TOS, (void *) &tos, sizeof(tos)))
	{
	    syslog(LOG_WARNING, "setsockopt(IPTOS_THROUGHPUT): %m");
	}
    }
#endif	
#ifdef TCP_NOPUSH
    if (setsockopt(fp->data_fd,
		   IPPROTO_TCP, TCP_NOPUSH, (void *) &one, sizeof(one)))
    {
	syslog(LOG_WARNING, "setsockopt(TCP_NOPUSH): %m");
    }
#endif
    
    fdata = a_malloc(sizeof(*fdata));
    fdata->fp = fp;
    fdata->handler = handler;
    fdata->vp = vp;
    
    pthread_attr_init(&cattr_detached);
    pthread_attr_setdetachstate(&cattr_detached,
				PTHREAD_CREATE_DETACHED);
    
    err = pthread_create(&tid, &cattr_detached,
			 ftpdata_thread, (void *) fdata);
    if (err)
    {
	syslog(LOG_ERR, "ftpdata_run: pthread_create: %s",
	       strerror(err));

	if (debug)
	    fprintf(stderr, "ftpdata_run: pthread_create(): failed\n");
	
	a_free(fdata);
	
	pthread_mutex_lock(&fp->data_mtx);
	s_close(fp->data_fd);
	fp->data_fd = -1;
	pthread_mutex_unlock(&fp->data_mtx);
    }
    pthread_attr_destroy(&cattr_detached);
 
    return 0;
}


void
ftpdata_wait(FTPCLIENT *fp)
{
    pthread_mutex_lock(&fp->data_mtx);
    while (fp->data_fd != -1)
	pthread_cond_wait(&fp->data_cv, &fp->data_mtx);
    pthread_mutex_unlock(&fp->data_mtx);
}


void
ftpdata_abort(FTPCLIENT *fp)
{
    if (debug)
	fprintf(stderr, "FTPDATA aborted\n");
    
    pthread_mutex_lock(&fp->data_mtx);
    if (fp->data_fd != -1)
    {
	shutdown(fp->data_fd, 2);
	fp->data_fd = -1;
    }
    pthread_mutex_unlock(&fp->data_mtx);
}

==> pftpd-0.37.tar.gz: pftpd-0.37/src/ftpdata.c <==
/*
** ftpdata.c - Manage the data connection
**
** Copyright (c) 1999-2000 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "config.h"

#include <stdio.h>
#include <syslog.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "pftpd.h"

#include "plib/aalloc.h"
#include "plib/threads.h"
#include "plib/safeio.h"
#include "plib/safestr.h"
#include "plib/support.h"
#include "plib/timeout.h"

#define PASV_TIMEOUT 120

static void *
ftpdata_thread(void *vp)
{
    FTPDATA *fdata = (FTPDATA *) vp;
    int err, code;
    


    if (debug)
	fprintf(stderr, "ftpdata_thread: start\n");
    
    err = fdata->handler(fdata->fp, fdata->vp);

    if (debug)
	fprintf(stderr, "ftpdata_thread: handler called\n");
    
    pthread_mutex_lock(&fdata->fp->data_mtx);
    if ((code = s_close(fdata->fp->data_fd)) < 0 && debug)
	fprintf(stderr, "close(%d): %s\n", fdata->fp->data_fd, strerror(errno));
	    
    fdata->fp->data_fd = -1;
    pthread_mutex_unlock(&fdata->fp->data_mtx);
    pthread_cond_broadcast(&fdata->fp->data_cv);
    
    switch (err)
    {
      case 0:
	break;
	
      case 226:
	fd_puts(fdata->fp->fd, "226 Transfer complete.\n");
	break;

      case 425:
	fd_puts(fdata->fp->fd, "425 Can not build data connection.\n");
	break;

      case 426:
	fd_puts(fdata->fp->fd, "426 Connection closed; transfer aborted.\n");
	break;

      default:
	fd_printf(fdata->fp->fd, "%d ???\n");
    }
    
    a_free(fdata);
    if (debug)
	fprintf(stderr, "ftpdata_thread: Stop\n");
    
    return NULL;

}


#if 0
static void
timeout_handler(void *vp)
{
    int fd = * (int *) vp;

    write(1, "TIMEOUT\n", 8);

    /* XXX: This is not good, what happens if there is a race between
       this code and s_accept(). Fixme! */
    
    * (int *) vp = -1;
    s_close(fd);
}
#endif


int
ftpdata_run(FTPCLIENT *fp,
	    const char *reason,
	    int (*handler)(FTPCLIENT *fp, void *vp),
	    void *vp)
{
    int err, try;
    int one = 1;
    FTPDATA *fdata;
    pthread_t tid;
    struct sockaddr_gen sin;
    socklen_t slen;
    pthread_attr_t cattr_detached;
    
    
    if (debug)
	fprintf(stderr, "ftpdata_run(\"%s\"): start\n",
		reason);
    
    /* Already have an active FTPDATA session? -> Fail */
    pthread_mutex_lock(&fp->data_mtx);
    if (fp->data_fd != -1)
    {
	pthread_mutex_unlock(&fp->data_mtx);
	syslog(LOG_ERR, "Multiple active FTPDATA sessions not supported");
	return -1;
    }
    

    fd_printf(fp->fd,
	      "150 Opening %s mode data connection for %s.\n",
	      fp->type == ftp_binary ? "BINARY" : "ASCII",
	      reason);

    if (fp->pasv != -1)
    {
	/* Passive connection */

	if (debug)
	    fprintf(stderr, "Waiting for PASV connection\n");
	
	slen = sizeof(sin);
	fp->data_fd = s_accept2(fp->pasv, (struct sockaddr *) &sin, &slen,
				PASV_TIMEOUT);
	pthread_mutex_unlock(&fp->data_mtx);
	
	if (fp->data_fd < 0)
	{
	    if (fp->pasv != -1)
	    {
		s_close(fp->pasv);
		fp->pasv = -1;
	    }
	    return -1;
	    
	    syslog(LOG_ERR,
		   "accept() failed when waiting for Passive FTPDATA connection");
	}

	
	if (debug)
	{
	    char buf1[256];
	    
	    fprintf(stderr, "Got passive connection from:\n");
	    fprintf(stderr, "\tsin_family = %d\n", SGFAM(sin));
	    fprintf(stderr, "\tsin_addr:sin_port = %s\n",
		    s_inet_ntox(&sin, buf1, sizeof(buf1)));
	}
	
	/* Make sure the connection came from the same host as the
	   control connection was from */

	if (SGSIZE(sin) != SGSIZE(fp->cp->rsin) ||
	    memcmp(SGADDRP(sin),
		   SGADDRP(fp->cp->rsin),
		   SGSIZE(sin)) != 0)
	{
	    char buf1[256], buf2[256];
	    
	    syslog(LOG_WARNING,
		   "%s - received foreign PASV connection from %s\n",
		   s_inet_ntox(&fp->cp->rsin, buf1, sizeof(buf1)),
		   s_inet_ntox(&sin, buf2, sizeof(buf2)));

	    pthread_mutex_unlock(&fp->data_mtx);
	    s_close(fp->data_fd);
	    fp->data_fd = -1;
	    pthread_mutex_unlock(&fp->data_mtx);
	    fd_puts(fp->fd, "426 Connection closed; transfer aborted.\n");
	    return 0;
	}

	s_close(fp->pasv);
	fp->pasv = -1;
    }
    else
    {
	/* Active connection */
	
	if (!SGPORT(fp->port))
	{
	    pthread_mutex_unlock(&fp->data_mtx);
	    syslog(LOG_ERR, "Sanity check - result: Total insanity");
	    return -1;
	}
	
	SGINIT(sin);
	slen = SGSOCKSIZE(sin);

	/* Get the local address that the FTP client connected to */
	if (getsockname(fp->cp->fd, (struct sockaddr *) &sin, &slen) != 0)
	{
	    pthread_mutex_unlock(&fp->data_mtx);
	    syslog(LOG_WARNING, "getsockname: %m");
	    return -1;
	}

	fp->data_fd = rpa_get_socket(rpap, &sin);
	if (fp->data_fd < 0)
	{
	    fp->data_fd = socket(SGFAM(sin), SOCK_STREAM, 0);
	    pthread_mutex_unlock(&fp->data_mtx);
	    
	    if (fp->data_fd < 0)
	    {
		syslog(LOG_ERR, "Failed allocating socket() for Active FTPDATA session");
		return -1;
	    }
	    
	    if (setsockopt(fp->data_fd,
			   SOL_SOCKET, SO_REUSEADDR, (void *) &one, sizeof(one)))
	    {
		syslog(LOG_WARNING, "setsockopt(SO_REUSEADDR): %m");
	    }
	    
	    SGPORT(sin) = 0; /* Any available source port */
	    
	    try = 0;
	    while (s_bind(fp->data_fd, (struct sockaddr *) &sin, SGSOCKSIZE(sin)) < 0)
	    {
		char buf1[256];
		
		if (debug)
		    fprintf(stderr,
			    "ftpdata_run: s_bind(): try=%d, addr=%s, failed: %s\n",
			    try,
			    s_inet_ntox(&sin, buf1, sizeof(buf1)),
			    strerror(errno));
		
		if (try++ > 10 || errno != EADDRINUSE)
		{
		    char buf1[256];
		    
		    
		    syslog(LOG_WARNING, "bind(%d,%s): %m",
			   fp->data_fd,
			   s_inet_ntox(&sin, buf1, sizeof(buf1)));
		    
		    pthread_mutex_lock(&fp->data_mtx);
		    s_close(fp->data_fd);
		    fp->data_fd = -1;
		    pthread_mutex_unlock(&fp->data_mtx);
		    return -1;
		}
		sleep(try);
	    }
	}
	else
	    pthread_mutex_unlock(&fp->data_mtx);
	
	if (connect(fp->data_fd,
		    (struct sockaddr *) &fp->port,
		    SGSOCKSIZE(fp->port)) < 0)
	{
	    char buf1[256];
	    
	    
	    syslog(LOG_WARNING, "connect(%d,%s): %m",
		   fp->data_fd,
		   s_inet_ntox(&fp->port, buf1, sizeof(buf1)));
	    
	    pthread_mutex_lock(&fp->data_mtx);
	    s_close(fp->data_fd);
	    fp->data_fd = -1;
	    pthread_mutex_unlock(&fp->data_mtx);
	    return -1;
	}
    }
    
    if (setsockopt(fp->data_fd,
		   IPPROTO_TCP, TCP_NODELAY, (void *) &one, sizeof(one)))
    {
	syslog(LOG_WARNING, "setsockopt(TCP_NODELAY): %m");
    }
    
#ifdef IPTOS_THROUGHPUT
    {
	int tos = IPTOS_THROUGHPUT;
	if (setsockopt(fp->data_fd,
		       IPPROTO_IP, IP_TOS, (void *) &tos, sizeof(tos)))
	{
	    syslog(LOG_WARNING, "setsockopt(IPTOS_THROUGHPUT): %m");
	}
    }
#endif	
#ifdef TCP_NOPUSH
    if (setsockopt(fp->data_fd,
		   IPPROTO_TCP, TCP_NOPUSH, (void *) &one, sizeof(one)))
    {
	syslog(LOG_WARNING, "setsockopt(TCP_NOPUSH): %m");
    }
#endif
    
    fdata = a_malloc(sizeof(*fdata));
    fdata->fp = fp;
    fdata->handler = handler;
    fdata->vp = vp;
    
    pthread_attr_init(&cattr_detached);
    pthread_attr_setdetachstate(&cattr_detached,
				PTHREAD_CREATE_DETACHED);

    err = pthread_create(&tid, &cattr_detached,
			 ftpdata_thread, (void *) fdata);
    if (err)
    {
	syslog(LOG_ERR, "ftpdata_run: pthread_create: %s",
	       strerror(err));

	if (debug)
	    fprintf(stderr, "ftpdata_run: pthread_create(): failed\n");
	
	a_free(fdata);
	
	pthread_mutex_lock(&fp->data_mtx);
	s_close(fp->data_fd);
	fp->data_fd = -1;
	pthread_mutex_unlock(&fp->data_mtx);
    }
    pthread_attr_destroy(&cattr_detached);
 
    return 0;
}


void
ftpdata_wait(FTPCLIENT *fp)
{
    pthread_mutex_lock(&fp->data_mtx);
    while (fp->data_fd != -1)
	pthread_cond_wait(&fp->data_cv, &fp->data_mtx);
    pthread_mutex_unlock(&fp->data_mtx);
}


void
ftpdata_abort(FTPCLIENT *fp)
{
    if (debug)
	fprintf(stderr, "FTPDATA aborted\n");
    
    pthread_mutex_lock(&fp->data_mtx);
    if (fp->data_fd != -1)
    {
	shutdown(fp->data_fd, 2);
	fp->data_fd = -1;
    }
    pthread_mutex_unlock(&fp->data_mtx);
}
